# -*- coding: utf-8 -*-

localization = {
    'global.quitConfirmation': u'Do you really want to quit?',
    'global.ok': u'OK',
    'global.quit': u'Quit',
    'global.cancel': u'Cancel',
    'global.saveFile': u'Save file',
    'cocoa.menu.about': u'About',
    'cocoa.menu.services': u'Services',
    'cocoa.menu.view': u'View',
    'cocoa.menu.hide': u'Hide',
    'cocoa.menu.hideOthers': u'Hide Others',
    'cocoa.menu.showAll': u'Show All',
    'cocoa.menu.quit': u'Quit',
    'cocoa.menu.fullscreen': u'Enter Fullscreen',
    'windows.fileFilter.allFiles': u'All files',
    'windows.fileFilter.otherFiles': u'Other file types',
    'linux.openFile': u'Open file',
    'linux.openFiles': u'Open files',
    'linux.openFolder': u'Open folder',
}
